/*
 * Framebuffer driver for VBE 2.0+ compliant graphic boards - kernel thread 
 * and vm86 routines.
 *
 * This code has to be compiled into the kernel even if vesafb is configured
 * as a module. If vesafb_thread were to be started while the module is being 
 * initialized, it would share its active_mm with modprobe. This mm would be 
 * lost after modprobe finished its work, and we can't allow it, because we 
 * need it for as long as the vesafb thread is active. 
 *
 * (c) 2004-2005 Michal Januszewski <spock@gentoo.org>
 * 
 */

#include <linux/workqueue.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/signal.h>
#include <linux/suspend.h>
#include <linux/unistd.h>
#include <video/vesa.h>
#include <video/edid.h>
#include <asm/mman.h>
#include <asm/page.h>
#include <asm/vm86.h>
#include <asm/thread_info.h>
#include <asm/uaccess.h>
#include "edid.h"

#ifdef DEBUG
#define DPRINTK(fmt, args...)	printk(KERN_DEBUG "%s: " fmt, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

int			vesafb_pid = 0;
struct vm86_struct	vm86;

static DECLARE_MUTEX(vesafb_sem);
static LIST_HEAD(vesafb_task_list);
static DECLARE_WAIT_QUEUE_HEAD(vesafb_wait);

_syscall3(int,ioperm,unsigned long, a, unsigned long, b, unsigned long, c);
_syscall1(int,vm86old,struct vm86_struct __user*, v86);

#define DEFAULT_VM86_FLAGS (IF_MASK | IOPL_MASK) 
#define VM86_PUSHW(x)	vm86.regs.esp -= 2; *(unsigned short*)(real_mem+vm86.regs.esp) = x;
#define REAL_MEM_SIZE	0x4000
#define REAL_MEM	0x2000
#define RET_CODE_SIZE 	0x02
#define STACK_SIZE	0x500
#define BUFFER		(STACK_SIZE + RET_CODE_SIZE)
#define FLAG_D 		(1 << 10)

/* segment prefix opcodes */
enum {	
	P_CS = 0x2e, 
	P_SS = 0x36, 
	P_DS = 0x3e,
	P_ES = 0x26, 
	P_FS = 0x64, 
	P_GS = 0x65
};

void vesafb_queue_task(struct vesafb_task *task)
{
	list_add_tail(&task->node, &vesafb_task_list);
	wake_up(&vesafb_wait);
}

/* emulated vm86 ins instruction */
static void vm86_ins(int size)
{
	u32 edx, edi;

	edx = vm86.regs.edx & 0xffff;
	edi = (vm86.regs.edi & 0xffff) + (u32)(vm86.regs.es << 4);

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("std\n");
	else	
		asm volatile ("cld\n");
	
	switch (size) {
	case 4:	asm volatile ("insl\n" : "=D" (edi) : "d" (edx), "0" (edi)); break;
	case 2:	asm volatile ("insw\n" : "=D" (edi) : "d" (edx), "0" (edi)); break;
	case 1:	asm volatile ("insb\n" : "=D" (edi) : "d" (edx), "0" (edi));
	}

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("cld\n");

	edi -= (u32)(vm86.regs.es << 4);

	vm86.regs.edi &= 0xffff0000;
	vm86.regs.edi |= edi & 0xffff;
}

static void vm86_rep_ins(int size)
{
	u16 cx = vm86.regs.ecx; 
	
	while (cx--)
		vm86_ins(size);

	vm86.regs.ecx &= 0xffff0000;
}

/* emulated vm86 outs instruction */
static void vm86_outs(int size, int segment)
{
	u32 edx, esi, base;

	edx = vm86.regs.edx & 0xffff;
	esi = vm86.regs.esi & 0xffff;

	switch (segment) {
	case P_CS: base = vm86.regs.cs; break;
	case P_SS: base = vm86.regs.ss; break;
	case P_ES: base = vm86.regs.es; break;
	case P_FS: base = vm86.regs.fs; break;
	case P_GS: base = vm86.regs.gs; break;
	default:   base = vm86.regs.ds; break;
	}

	esi += base << 4;

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("std\n");
	else
		asm volatile ("cld\n");

	switch (size) {
	case 4: asm volatile ("outsl\n" : "=S" (esi) : "d" (edx), "0" (esi)); break;
	case 2: asm volatile ("outsw\n" : "=S" (esi) : "d" (edx), "0" (esi)); break;
	case 1: asm volatile ("outsb\n" : "=S" (esi) : "d" (edx), "0" (esi)); break;
	}

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("cld");
	
	esi -= base << 4;
	vm86.regs.esi &= 0xffff0000;
	vm86.regs.esi |= (esi & 0xffff);
}

static void vm86_rep_outs(int size, int segment)
{
	u16 cx = vm86.regs.ecx;
	
	while (cx--)
		vm86_outs(size, segment);

	vm86.regs.ecx &= 0xffff0000;
}

void vesafb_do_vm86(struct vm86_regs *regs)
{
        unsigned char *real_mem = (void*)REAL_MEM;
	unsigned int ret;
	
	memset(&vm86,0,sizeof(vm86));
	memcpy(&vm86.regs, regs, sizeof(struct vm86_regs));

	/* the return code */
	real_mem[0] = 0xcd;  		/* int opcode */
	real_mem[1] = 0xff;		/* int number (255) */

        /* we use int 255 to get back to protected mode */
	memset(&vm86.int_revectored, 0, sizeof(vm86.int_revectored));
        ((unsigned char *) &vm86.int_revectored)[0xff / 8] |= (1 << (0xff % 8));	/* int 0xff */
	
	/* it's up to the caller to set the rest of the registers */
	vm86.regs.eflags = DEFAULT_VM86_FLAGS;
	vm86.regs.cs = *(unsigned short *)0x42;		/* 0x10 * 4 + 2 - the int map starts at 0x0	*/
	vm86.regs.eip = *(unsigned short *)0x40;	/* 0x10 * 4					*/
	
	/* stack @ 0x10500, size: 0x500-4 - should be enough for our needs */
	vm86.regs.ss = (REAL_MEM >> 4);
	vm86.regs.esp = STACK_SIZE+RET_CODE_SIZE;

	/* these will be fetched off the stack when we come to an iret in the int's 0x10 code */
	VM86_PUSHW(DEFAULT_VM86_FLAGS);
	VM86_PUSHW((REAL_MEM >> 4));		/* return code segment */
	VM86_PUSHW(0x0000);			/* return code offset */

	while(1) {
		ret = vm86old(&vm86);

		if (VM86_TYPE(ret) == VM86_INTx) {

			int vint = VM86_ARG(ret);

			/* if exit from vm86 was caused by int 0xff - we're done.. */
			if (vint == 0xff)
				goto vm86_done_call;
			
			/* .. otherwise, we have call the int handler manually */
			VM86_PUSHW(vm86.regs.eflags);
			VM86_PUSHW(vm86.regs.cs);
			VM86_PUSHW(vm86.regs.eip);

			vm86.regs.cs = *(unsigned short *)((vint << 2) + 2);
			vm86.regs.eip = *(unsigned short *)(vint << 2);
			vm86.regs.eflags &= ~(VIF_MASK | TF_MASK);

		} else if (VM86_TYPE(ret) == VM86_UNKNOWN) {
			
			u8 *instr;
			u8 data32 = 0, segment = P_DS, rep = 0;
			int i = 0;
					
			instr = (u8*)((vm86.regs.cs << 4) + vm86.regs.eip);

			while (1) {
			
				switch(instr[i]) {

				case 0x66:	/* operand size prefix */
					data32 = 1 - data32; 
					i++;
					break;
				case 0xf2:	/* repnz */
				case 0xf3:	/* rep */
					rep = 1;
					i++;
					break;
				case P_CS:	/* segment prefix */
				case P_SS:
				case P_DS:
				case P_ES:
				case P_FS:
				case P_GS:
					segment = instr[i];
					i++;
					break;
				case 0xf0:	/* LOCK - ignored */
				case 0x67:	/* address size prefix - ignored */
					i++;
					break;
				case 0x6c:	/* insb */
					if (rep)	vm86_rep_ins(1);
					else		vm86_ins(1);
					i++;
					goto vm86_done_emu;
				case 0x6d:	/* insw / insd */
					if (rep)
						if (data32)	vm86_rep_ins(4);
						else		vm86_rep_ins(2);
					else
						if (data32)	vm86_ins(4);
						else		vm86_ins(2);
					i++;
					goto vm86_done_emu;
				case 0x6e:	/* outsb */
					if (rep)	vm86_rep_outs(1, segment);
					else		vm86_outs(1, segment);
					i++;
					goto vm86_done_emu;
				case 0x6f:	/* outsw / outsd */
					if (rep)
						if (data32)	vm86_rep_outs(4, segment);
						else		vm86_rep_outs(2, segment);
					else
						if (data32)	vm86_outs(4, segment);
						else		vm86_outs(2, segment);
					i++;
					goto vm86_done_emu;
				case 0xe4:	/* inb xx */
					asm volatile (
						"inb %w1, %b0" 
						: "=a" (vm86.regs.eax)
						: "d" (instr[i+1]), "0" (vm86.regs.eax));
					i += 2;
					goto vm86_done_emu;
				case 0xe5:	/* inw xx / ind xx */
					if (data32)
						asm volatile (
							"inl %w1, %0"
							: "=a" (vm86.regs.eax)
							: "d" (instr[i+1]), "0" (vm86.regs.eax));
					else
						asm volatile (
							"inw %w1, %w0"
							: "=a" (vm86.regs.eax)
							: "d" (instr[i+1]), "0" (vm86.regs.eax));
					i += 2;
					goto vm86_done_emu;
				case 0xec:	/* inb dx */
					asm volatile (
						"inb %w1, %b0"
	 					: "=a" (vm86.regs.eax)
						: "d" (vm86.regs.edx), "0" (vm86.regs.eax));
					i++;
					goto vm86_done_emu;
				case 0xed:	/* inw dx / ind dx */
					if (data32)
						asm volatile (
							"inl %w1, %0"
							: "=a" (vm86.regs.eax)
							: "d" (vm86.regs.edx));
					else
						asm volatile (
							"inw %w1, %w0"
							: "=a" (vm86.regs.eax)
							: "d" (vm86.regs.edx));
					i++;
					goto vm86_done_emu;
				case 0xe6:	/* outb xx */
					asm volatile (
						"outb %b0, %w1"
						: : "a" (vm86.regs.eax), "d" (instr[i+1]));
					i += 2;
					goto vm86_done_emu;
				case 0xe7:	/* outw xx / outd xx */
					if (data32)
						asm volatile (
							"outl %0, %w1"
							: : "a" (vm86.regs.eax), "d" (instr[i+1]));
					else
						asm volatile (
							"outw %w0, %w1"
							: : "a" (vm86.regs.eax), "d" (instr[i+1]));
					i += 2;
					goto vm86_done_emu;
				case 0xee:	/* outb dx */
					asm volatile (
						"outb %b0, %w1"
						: : "a" (vm86.regs.eax), "d" (vm86.regs.edx));
					i++;
					goto vm86_done_emu;
				case 0xef:	/* outw dx / outd dx */
					if (data32)
						asm volatile (
							"outl %0, %w1"
							: : "a" (vm86.regs.eax), "d" (vm86.regs.edx));
					else
						asm volatile (
							"outw %w0, %w1"
							: : "a" (vm86.regs.eax), "d" (vm86.regs.edx));
					i++;
					goto vm86_done_emu;
				default:
					printk(KERN_ERR "vesafb: BUG, opcode %x emulation not supported\n", instr[i]);
					goto vm86_done_call;
				}
			}
vm86_done_emu:		vm86.regs.eip += i;
		} else {
			printk(KERN_ERR "vesafb: BUG, returned from vm86 with %x\n", ret);
			goto vm86_done_call;
		}	
	}

vm86_done_call:

	/* copy the registers' state back to the caller's struct */
	memcpy(regs, &vm86.regs, sizeof(struct vm86_regs));
}

#define vesafb_get_string(str) { \
													\
	/* the address is in the form ssssoooo, where oooo = offset, ssss = segment */			\
	addr = ((vbe_pib(task->res)->str & 0xffff0000) >> 12) +						\
		(vbe_pib(task->res)->str & 0x0000ffff);							\
													\
	/* the data is in ROM which is shared between processes, so we just translate the		\
	   real mode address into one visible from the kernel space */					\
	if (addr >= 0xa0000) {										\
		vbe_pib(task->res)->str = (u32) __va(addr);						\
													\
	/* the data is in the buffer, we just have to convert the address so that it would		\
	   point into the buffer user provided */							\
	} else if (addr > REAL_MEM+BUFFER && addr < REAL_MEM+BUFFER + 					\
		   sizeof(struct vesafb_vbe_info_block)) {						\
		addr -= BUFFER+REAL_MEM;					 			\
		vbe_pib(task->res)->str = (u32) (task->res + addr);					\
													\
	/* this should never happen: someone was insane enough to put the data somewhere in the RAM */	\
	} else {											\
		vbe_pib(task->res)->str = (u32) "";							\
	}												\
}

void vesafb_handle_tasks(void)
{
	struct vesafb_task *task;
	struct list_head *node, *next;
	int addr, res;
	
	down(&vesafb_sem);
	list_for_each_safe(node, next, &vesafb_task_list) {

		task = container_of(node, struct vesafb_task, node);  
		
		switch (task->type) {

			case VESAFB_TASK_DOVM86:
				vesafb_do_vm86(&task->regs);
				break;
				
			case VESAFB_TASK_GETVBE_IB:
				task->regs.es	= (REAL_MEM >> 4);
				task->regs.edi	= BUFFER;
				strncpy(vbe_pib(REAL_MEM+BUFFER)->vbe_signature,"VBE2",4);
				
				vesafb_do_vm86(&task->regs);

				memcpy(task->res, (void*)(REAL_MEM + BUFFER), sizeof(struct vesafb_vbe_info_block));
			
				/* the OEM fields were not defined prior to VBE 2.0 */
				if (vbe_pib(task->res)->vbe_version >= 0x200) {
					vesafb_get_string(oem_string_ptr);
					vesafb_get_string(oem_vendor_name_ptr);
					vesafb_get_string(oem_product_name_ptr);
					vesafb_get_string(oem_product_rev_ptr);
				}
		
				/* this is basically the same as vesafb_get_string */
				addr = ((vbe_pib(task->res)->mode_list_ptr & 0xffff0000) >> 12) +
					(vbe_pib(task->res)->mode_list_ptr & 0x0000ffff);
			
				if (addr >= 0xa0000) {
					vbe_pib(task->res)->mode_list_ptr = (u32) __va(addr);
				} else if (addr > REAL_MEM+BUFFER && addr < REAL_MEM+BUFFER +
					   sizeof(struct vesafb_vbe_info_block))
				{
					addr -= BUFFER+REAL_MEM;
					vbe_pib(task->res)->mode_list_ptr = (u32) (task->res + addr);
				} else {
					res = 0;
					printk(KERN_WARNING "vesafb: warning, copying modelist from somewhere in RAM!\n");
					while (*(u16*)(addr+res) != 0xffff && 
					       res < (sizeof(vbe_pib(task->res)->reserved) - 2) )
					{
						*(u16*) ((u32)&(vbe_pib(task->res)->reserved) + res) =
							*(u16*)(addr+res);
						res += 2;
					}
					
					*(u16*) ((u32)&(vbe_pib(task->res)->reserved) + res) = 0xffff;
 				}
				break;

			case VESAFB_TASK_GETVBE_MODEINFO:
				task->regs.es	= (REAL_MEM >> 4);
				task->regs.edi	= BUFFER;
				vesafb_do_vm86(&task->regs);
				memcpy(task->res, (void*)(REAL_MEM + BUFFER), sizeof(struct vesafb_mode_info_block));
				break;

			case VESAFB_TASK_SWITCHMODE:
				if (task->res != NULL) {
					task->regs.es	= (REAL_MEM >> 4);
					task->regs.edi	= BUFFER;
					memcpy((void*)(REAL_MEM + BUFFER), task->res, sizeof(struct vesafb_crtc_info_block));
				}
				
				vesafb_do_vm86(&task->regs);
				break;

			case VESAFB_TASK_SETPAL:
				task->regs.es	= (REAL_MEM >> 4);
				task->regs.edi	= BUFFER;
				memcpy((void*)(REAL_MEM + BUFFER), task->res, sizeof(struct vesafb_pal_entry));
				vesafb_do_vm86(&task->regs);
				break;
				
			case VESAFB_TASK_GETEDID:
				task->regs.es	= (REAL_MEM >> 4);
				task->regs.edi	= BUFFER;
				vesafb_do_vm86(&task->regs);
				memcpy(task->res, (void*)(REAL_MEM + BUFFER), EDID_LENGTH);
			
			default:
				break;
		}
		
		task->done = 1;
		if (task->flags & VESAFB_FLAG_FREESTRUCT) {
			kfree(task);
		}
	}

	up(&vesafb_sem);
	list_del_init(node);
}

int vesafb_thread(void *unused)
{
	struct vm_area_struct vma;
	struct page *page;
	
	int ret, err = 0;
	void *mem;	

	set_fs(KERNEL_DS);
	daemonize("vesafb");
	DPRINTK("started vesafb thread\n");

	current->mm = current->active_mm;
	mem = kmalloc(REAL_MEM_SIZE,GFP_KERNEL);
	
	if (!mem)
		return -ENOMEM;

	for (page = virt_to_page(mem); page < virt_to_page(mem+REAL_MEM_SIZE); page++) {
		SetPageReserved(page);
        }

#ifdef CONFIG_SMP
	cpus_clear(current->active_mm->cpu_vm_mask);
	cpu_set(smp_processor_id(), current->active_mm->cpu_vm_mask);
#endif

	vma.vm_mm = current->active_mm;
	vma.vm_page_prot.pgprot = PROT_READ | PROT_EXEC | PROT_WRITE;
	
	ret = io_remap_page_range(&vma, 0x000000, __pa(mem), REAL_MEM_SIZE, vma.vm_page_prot);
	ret += io_remap_page_range(&vma, 0x0a0000, 0x0a0000, 0x100000 - 0x0a0000, vma.vm_page_prot);

	if (ret) {
		printk(KERN_ERR "vesafb thread: memory remapping failed\n");
		err = -EINVAL;
		goto thr_end;
	}

	/* copy the first 0x20000 bytes from low mem to our private memory, which is 
	 * then used for the vm86 calls */
	memcpy((void*)0x0, __va(0x0), REAL_MEM_SIZE);

	ioperm(0,1024,1);	/* we can live if it fails, so don't bother checking for errors */
	set_user_nice(current, -10);
	
	while (1) {
		vesafb_handle_tasks();
		wait_event_interruptible(vesafb_wait, !list_empty(&vesafb_task_list));
		try_to_freeze();
		if (signal_pending(current))
			break;
	}

thr_end:
	DPRINTK("exiting the vesafb thread\n");
	vesafb_pid = 0;
	
	for (page = virt_to_page(mem); page < virt_to_page(mem+REAL_MEM_SIZE); page++) {
		ClearPageReserved(page);
        }
	
	kfree(mem);
	return err;
}

int vesafb_init_thread(void)
{
	vesafb_pid = kernel_thread(vesafb_thread,NULL,0);
	return 0;
}

EXPORT_SYMBOL(vesafb_pid);
EXPORT_SYMBOL(vesafb_queue_task);
