/*
 * Framebuffer driver for VBE 2.0+ compliant graphic boards
 *
 * (c) 2004-2005 Michal Januszewski <spock@gentoo.org>
 *     Based upon vesafb code by Gerd Knorr <kraxel@goldbach.in-berlin.de>
 * 
 */

#ifdef DEBUG
#define DPRINTK(fmt, args...)	printk(KERN_DEBUG "%s: " fmt, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <video/edid.h>
#include <video/vesa.h>

#include <asm/io.h>
#include <asm/mtrr.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include "edid.h"

#define dac_reg	(0x3c8)
#define dac_val	(0x3c9)

#define VESAFB_NEED_EXACT_RES 1
#define VESAFB_NEED_EXACT_DEPTH 2

/* --------------------------------------------------------------------- */

static struct fb_var_screeninfo vesafb_defined __initdata = {
	.activate	= FB_ACTIVATE_NOW,
	.height		= -1,
	.width		= -1,
	.right_margin	= 32,
	.upper_margin	= 16,
	.lower_margin	= 4,
	.vsync_len	= 4,
	.vmode		= FB_VMODE_NONINTERLACED,
};

static struct fb_fix_screeninfo vesafb_fix __initdata = {
	.id	= "VESA VGA",
	.type	= FB_TYPE_PACKED_PIXELS,
	.accel	= FB_ACCEL_NONE,
};

static int				mtrr       = 1;	/* use MTRR */
static int				ypan       = 0;	/* 0 - nothing, 1 - ypan, 2 -ywrap */
static int				pmi_setpal = 0;	/* pmi for palette changes */
static unsigned short			*pmi_base  = NULL; /* protected mode interface location in memory */
static void				(*pmi_start)(void) = NULL;	
static void				(*pmi_pal)(void) = NULL;
static struct task_struct		*vesafb_serv_thread = NULL;
static struct vesafb_vbe_info_block	vbe_ib;
static struct vesafb_mode_info_block	*vbe_modes;
static int				vbe_modes_cnt = 0;
static u8				mon_limits = 0;	/* 0 - no monitor limits, 1 - full monitor limits, 
							   2 - monitor limits with default pixel clock */
static int				nocrtc = 0;	/* ignore CRTC settings */
static struct fb_info			*vesafb_info = NULL;
static struct fb_videomode		*edid_modes    __initdata = NULL;
static int				edid_modes_cnt __initdata = 0; 
static int				noedid         __initdata = 0; /* don't try the DDC transfers */
static int				vram_remap     __initdata = 0; /* set amount of memory to be used */
static int				vram_total     __initdata = 0; /* set total amount of memory */
static unsigned short			maxclk         __initdata = 0; /* maximum pixel clock */
static unsigned short			maxvf          __initdata = 0; /* maximum vertical frequency */
static unsigned short			maxhf          __initdata = 0; /* maximum horizontal frequency */
static int 				gtf            __initdata = 0; /* forces use of the GTF */
static char				*mode_option   __initdata = NULL;
static unsigned short			vbemode	                  = 0;

extern int vesafb_pid;			/* PID of the vesafb service thread */

/* --------------------------------------------------------------------- */

#define vesafb_create_task(task)	{ task = kmalloc(sizeof(struct vesafb_task), GFP_ATOMIC); \
					  if (task) memset(task,0,sizeof(struct vesafb_task));	}

#define vesafb_wait_for_task(task)	{ while (task->done == 0) { schedule(); } }

extern void vesafb_queue_task(struct vesafb_task *task);
static int vesafb_find_vbe_mode(int xres, int yres, int bpp, unsigned char flags);

struct fb_videomode *fb_create_modedb(unsigned char *edid, int *dbsize);
int fb_get_monitor_limits(unsigned char *edid, struct fb_monspecs *specs);

/* --------------------------------------------------------------------- */

static int vesafb_pan_display(struct fb_var_screeninfo *var,
                              struct fb_info *info)
{
#ifdef __i386__
	int offset;
	
	if (!ypan)
		return -EINVAL;
	if (var->xoffset)
		return -EINVAL;
	if (var->yoffset > var->yres_virtual)
		return -EINVAL;
	if ((ypan==1) && var->yoffset+var->yres > var->yres_virtual)
		return -EINVAL;

	offset = (var->yoffset * info->fix.line_length + var->xoffset) / 4;

	/* It turns out it's not the best idea to do panning via vm86,
	 * so we only allow it if we have a PMI. */
	if (pmi_start) {
		__asm__ __volatile__(
			"call *(%%edi)"
			: /* no return value */
			: "a" (0x4f07),         /* EAX */
			  "b" (0),              /* EBX */
			  "c" (offset),         /* ECX */
			  "d" (offset >> 16),   /* EDX */
			  "D" (&pmi_start));    /* EDI */
	}
#endif
	return 0;
}

static void vesa_setpalette(int regno, unsigned red, unsigned green, unsigned blue, int shift)
{
	struct vesafb_pal_entry entry;
	struct vesafb_task *mytask;

#ifdef __i386__
	entry.red   = red   >> shift;
	entry.green = green >> shift;
	entry.blue  = blue  >> shift;
	entry.pad   = 0;

	if (pmi_setpal) {
	        __asm__ __volatile__(
                "call *(%%esi)"
                : /* no return value */
                : "a" (0x4f09),         /* EAX */
                  "b" (0),              /* EBX */
                  "c" (1),              /* ECX */
                  "d" (regno),          /* EDX */
                  "D" (&entry),         /* EDI */
                  "S" (&pmi_pal));      /* ESI */
	} else {
		vesafb_create_task (mytask);

		mytask->regs.eax = 0x4f09;
		mytask->regs.ebx = 0x0;
		mytask->regs.ecx = 1;
		mytask->regs.edx = regno;
		mytask->res = &entry;
		mytask->type = VESAFB_TASK_SETPAL;
			
		vesafb_queue_task (mytask);
		vesafb_wait_for_task(mytask);
		kfree(mytask);
	}
#endif
}

static int vesafb_setcolreg(unsigned regno, unsigned red, unsigned green,
			    unsigned blue, unsigned transp,
			    struct fb_info *info)
{
	/*
	 *  Set a single color register. The values supplied are
	 *  already rounded down to the hardware's capabilities
	 *  (according to the entries in the `var' structure). Return
	 *  != 0 for invalid regno.
	 */
	if (regno >= info->cmap.len)
		return 1;

	switch (info->var.bits_per_pixel) {
	case 8:
		vesa_setpalette(regno,red,green,blue,16 - info->var.green.length);
		break;
	case 16:
		if (info->var.red.offset == 10) {
			/* 1:5:5:5 */
			((u32*) (info->pseudo_palette))[regno] =	
					((red   & 0xf800) >>  1) |
					((green & 0xf800) >>  6) |
					((blue  & 0xf800) >> 11);
		} else {
			/* 0:5:6:5 */
			((u32*) (info->pseudo_palette))[regno] =	
					((red   & 0xf800)      ) |
					((green & 0xfc00) >>  5) |
					((blue  & 0xf800) >> 11);
		}
		break;
	case 24:
		red   >>= 8;
		green >>= 8;
		blue  >>= 8;
		((u32 *)(info->pseudo_palette))[regno] =
			(red   << info->var.red.offset)   |
			(green << info->var.green.offset) |
			(blue  << info->var.blue.offset);
		break;
	case 32:
		red   >>= 8;
		green >>= 8;
		blue  >>= 8;
		((u32 *)(info->pseudo_palette))[regno] =
			(red   << info->var.red.offset)   |
			(green << info->var.green.offset) |
			(blue  << info->var.blue.offset);
		break;
    }
    return 0;
}

static int vesafb_set_par(struct fb_info *info)
{
	struct vesafb_par *par = (struct vesafb_par *) info->par;
	struct vesafb_task *mytask;
	struct vesafb_crtc_info_block *crtc = NULL;	
	struct vesafb_mode_info_block *mode = NULL;
	int err = 0, i;

	/* Has the VBE mode number been specified? */
	if (vbemode > 0) {
		for (i = 0; i < vbe_modes_cnt; i++) {
			if (vbe_modes[i].mode_id == vbemode) {
				mode = &vbe_modes[i];
				break;
			}
		}
	} else {
		i = vesafb_find_vbe_mode(info->var.xres, info->var.yres, 
					 info->var.bits_per_pixel, VESAFB_NEED_EXACT_RES | VESAFB_NEED_EXACT_DEPTH);
		if (i > 0)
			mode = &vbe_modes[i];
	}

	if (mode == NULL)
		return -EINVAL;
	
	vesafb_create_task (mytask);

	mytask->regs.eax = 0x4f02;
	mytask->regs.ebx = mode->mode_id | 0x4000;	/* use LFB */

	if (vbe_ib.vbe_version >= 0x0300 && !nocrtc && !vbemode) {

		mytask->regs.ebx |= 0x0800; /* use CRTC data */
		crtc = kmalloc(sizeof(struct vesafb_crtc_info_block), GFP_KERNEL);
	
		if (!crtc) {
			err = -ENOMEM;
			goto out;
		}
		crtc->horiz_start = info->var.xres + info->var.right_margin;
		crtc->horiz_end	  = crtc->horiz_start + info->var.hsync_len;
		crtc->horiz_total = crtc->horiz_end + info->var.left_margin;
		
		crtc->vert_start  = info->var.yres + info->var.lower_margin;
		crtc->vert_end    = crtc->vert_start + info->var.vsync_len;
		crtc->vert_total  = crtc->vert_end + info->var.upper_margin;
	
		crtc->pixel_clock = PICOS2KHZ(info->var.pixclock) * 1000;
		crtc->refresh_rate = (u16)(100 * (crtc->pixel_clock / (crtc->vert_total * crtc->horiz_total)));
		crtc->flags = 0;

		if (info->var.vmode & FB_VMODE_DOUBLE)
			crtc->flags |= 0x1;

		if (info->var.vmode & FB_VMODE_INTERLACED)
			crtc->flags |= 0x2;
	
		if (!(info->var.sync & FB_SYNC_HOR_HIGH_ACT))
			crtc->flags |= 0x4;

		if (!(info->var.sync & FB_SYNC_VERT_HIGH_ACT))
			crtc->flags |= 0x8;
		
		memcpy(&par->crtc, crtc, sizeof(struct vesafb_crtc_info_block));
	} else
		memset(&par->crtc, 0, sizeof(struct vesafb_crtc_info_block));

	mytask->res = (void*)crtc;
	mytask->type = VESAFB_TASK_SWITCHMODE; 

	vesafb_queue_task (mytask);
	vesafb_wait_for_task(mytask);

	if ((mytask->regs.eax & 0xffff) != 0x004f) {
		printk(KERN_ERR "vesafb: mode switch failed (eax: 0x%lx)\n", mytask->regs.eax);
		err = -EINVAL;
		goto out;
	}

	/* The VBE mode number is valid only the first time the mode is set. */
	if (vbemode)
		vbemode = 0;
	
	if (vbe_ib.capabilities & VESAFB_CAP_CAN_SWITCH_DAC && mode->bits_per_pixel <= 8) {
		mytask->done = 0;
		mytask->type = VESAFB_TASK_DOVM86;
		mytask->regs.eax = 0x4f08;
		mytask->regs.ebx = 0x0800;
	
		vesafb_queue_task (mytask);
		vesafb_wait_for_task(mytask);

		if ((mytask->regs.eax & 0xffff) != 0x004f ||
		    ((mytask->regs.ebx & 0xff00) >> 8) != 8) {
			/* We've failed to set the DAC palette format - time to correct var. */
			info->var.red.length    = 6;
			info->var.green.length  = 6;
			info->var.blue.length   = 6;
		}
	}
		
	info->fix.visual = (info->var.bits_per_pixel == 8) ? FB_VISUAL_PSEUDOCOLOR : FB_VISUAL_TRUECOLOR;
	info->fix.line_length = mode->bytes_per_scan_line;
	par->vbe_mode = mode->mode_id;

	DPRINTK("set new mode %dx%d-%d (0x%x)\n", info->var.xres, info->var.yres, info->var.bits_per_pixel, mode->mode_id);

out:	if (crtc != NULL)
		kfree(crtc);
	kfree(mytask);

	return err;
}

static void vesafb_setup_var(struct fb_var_screeninfo *var, struct fb_info *info, 
			     struct vesafb_mode_info_block *mode)
{
	var->xres = mode->x_res;
	var->yres = mode->y_res;
	var->xres_virtual = mode->x_res;
	var->yres_virtual = (ypan) ? info->fix.smem_len / mode->bytes_per_scan_line : mode->y_res;
	var->xoffset = 0;
	var->yoffset = 0;
	var->bits_per_pixel = mode->bits_per_pixel;

	if (var->bits_per_pixel == 15)
		var->bits_per_pixel = 16;

	if (var->bits_per_pixel > 8) {
		var->red.offset    = mode->red_off;
		var->red.length    = mode->red_len;
		var->green.offset  = mode->green_off;
		var->green.length  = mode->green_len;
		var->blue.offset   = mode->blue_off;
		var->blue.length   = mode->blue_len;
		var->transp.offset = mode->rsvd_off;
		var->transp.length = mode->rsvd_len;

		DPRINTK("directcolor: size=%d:%d:%d:%d, shift=%d:%d:%d:%d\n",
	       		mode->rsvd_len,
			mode->red_len,
			mode->green_len,
			mode->blue_len,
			mode->rsvd_off,
			mode->red_off,
			mode->green_off,
			mode->blue_off);
	} else {
		var->red.offset    = 0;
		var->green.offset  = 0;
		var->blue.offset   = 0;
		var->transp.offset = 0;
		
		/* We're assuming that we can switch the DAC to 8 bits. If this proves
		 * to be incorrect, we'll update the fields later in set_par. */
		if (vbe_ib.capabilities & VESAFB_CAP_CAN_SWITCH_DAC) {
			var->red.length    = 8;
			var->green.length  = 8;
			var->blue.length   = 8;
			var->transp.length = 0;
		} else {
			var->red.length    = 6;
			var->green.length  = 6;
			var->blue.length   = 6;
			var->transp.length = 0;
		}
	}
}

static int inline vesafb_check_limits(struct fb_var_screeninfo *var, struct fb_info *info)
{	
	if (mon_limits)
		return fb_validate_mode(var, info);
	else
		return 0;
}

static int vesafb_find_vbe_mode(int xres, int yres, int bpp, unsigned char flags)
{
	int i, match = -1, h = 0, d = 0x7fffffff;
	
	for (i = 0; i < vbe_modes_cnt; i++) {

		h = abs(vbe_modes[i].x_res - xres) + abs(vbe_modes[i].y_res - yres) + 
		    abs(bpp - vbe_modes[i].bits_per_pixel);

		if (h == 0)
			return i;

		if (h < d || (h == d && vbe_modes[i].bits_per_pixel > bpp)) {
			d = h;
			match = i;
		}
	}

	i = 1;

	if (flags & VESAFB_NEED_EXACT_RES && d > 24)
		i = 0;
	
	if (flags & VESAFB_NEED_EXACT_DEPTH && d > 1)
		i = 0;
	
	if (i != 0)
		return match;
	else
		return -1;
}

static int vesafb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	int match = -1;
	
	if (vesafb_check_limits(var, info))
		return -EINVAL;
	
	/* FIXME: we should allow interlaced/double modes if an appropriate mode is supported by the VBE */
	if (var->vmode & (FB_VMODE_INTERLACED | FB_VMODE_DOUBLE))
		return -EINVAL;
	
	match = vesafb_find_vbe_mode(var->xres, var->yres, var->bits_per_pixel, VESAFB_NEED_EXACT_RES);

	if (match == -1) {
		DPRINTK("vesafb: mode %dx%d-%d not found\n", var->xres, var->yres, 
			var->bits_per_pixel);
		return -EINVAL;
	} else {
		var->bits_per_pixel = vbe_modes[match].bits_per_pixel;
		vesafb_setup_var(var, info, &vbe_modes[match]);
	
		DPRINTK("found mode 0x%x (%dx%d-%dbpp)\n",
			vbe_modes[match].mode_id, vbe_modes[match].x_res, vbe_modes[match].y_res, 
			vbe_modes[match].bits_per_pixel);
	}
	
	/* Check whether we have remapped enough memory for this mode.. */
	if (var->yres * vbe_modes[match].bytes_per_scan_line > info->fix.smem_len) {
		return -EINVAL;		
	}
	
	return 0;
}

static void vesafb_platform_release(struct device *device)
{
	return;
}

static int __init vesafb_probe(struct device *device);

static struct fb_ops vesafb_ops = {
	.owner		= THIS_MODULE,
	.fb_setcolreg	= vesafb_setcolreg,
	.fb_pan_display	= vesafb_pan_display,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_cursor	= soft_cursor,
	.fb_check_var	= vesafb_check_var,
	.fb_set_par	= vesafb_set_par
};

static struct device_driver vesafb_driver = {
	.name	= "vesafb",
	.bus	= &platform_bus_type,
	.probe	= vesafb_probe,
};

static struct platform_device vesafb_device = {
	.name	= "vesafb",
        .dev    = {
                .release = vesafb_platform_release,
        }
};

#ifndef MODULE
int __init vesafb_setup(char *options)
{
	char *this_opt;
	
	if (!options || !*options)
		return 0;
	
	DPRINTK("options %s\n",options);
	
	while ((this_opt = strsep(&options, ",")) != NULL) {
		if (!*this_opt) continue;
		
		DPRINTK("this_opt: %s\n",this_opt);
		
		if (! strcmp(this_opt, "redraw"))
			ypan=0;
		else if (! strcmp(this_opt, "ypan"))
			ypan=1;
		else if (! strcmp(this_opt, "ywrap"))
			ypan=2;
		else if (! strcmp(this_opt, "vgapal"))
			pmi_setpal=0;
		else if (! strcmp(this_opt, "pmipal"))
			pmi_setpal=1;
		else if (! strcmp(this_opt, "mtrr"))
			mtrr=1;
		else if (! strcmp(this_opt, "nomtrr"))
			mtrr=0;
		else if (! strcmp(this_opt, "nocrtc"))
			nocrtc=1;
		else if (! strcmp(this_opt, "noedid")) 
			noedid=1;
		else if (! strcmp(this_opt, "gtf"))
			gtf=1;
		else if (! strncmp(this_opt, "vtotal:", 7))
			vram_total = simple_strtoul(this_opt+7, NULL, 0);
		else if (! strncmp(this_opt, "vremap:", 7))
			vram_remap = simple_strtoul(this_opt+7, NULL, 0);
		else if (! strncmp(this_opt, "maxhf:", 6))
			maxhf = simple_strtoul(this_opt + 6, NULL, 0);
		else if (! strncmp(this_opt, "maxvf:", 6))
			maxvf = simple_strtoul(this_opt + 6, NULL, 0);
		else if (! strncmp(this_opt, "maxclk:", 7))
			maxclk = simple_strtoul(this_opt + 7, NULL, 0);
		else if (! strncmp(this_opt, "vbemode:", 8))
			vbemode = simple_strtoul(this_opt + 8, NULL,0);
		else if (this_opt[0] >= '0' && this_opt[0] <= '9') {
			DPRINTK("mode_option: %s\n",this_opt);
			mode_option = this_opt;
		} else {
			printk(KERN_WARNING "vesafb: unrecognized option %s\n", this_opt);
		}
	}

	return 0;
}
#endif /* !MODULE */

static int vesafb_read_proc_modes(char *buf, char **start, off_t offset,
			    	  int len, int *eof, void *private)
{
	int clen = 0, i;
	
	for (i = 0; i < vbe_modes_cnt; i++) {
		clen += sprintf(buf + clen, "%dx%d-%d\n", vbe_modes[i].x_res, 
				vbe_modes[i].y_res, vbe_modes[i].bits_per_pixel);
	}
	*start = buf + offset;

	if (clen > offset) {
		clen -= offset;
	} else {
		clen = 0;
	}
	return clen;
}

static int vesafb_read_proc_vbe_info(char *buf, char **start, off_t offset,
			    	     int len, int *eof, void *private)
{
	int clen = 0;

	clen += sprintf(buf + clen, "Version:    %d.%d\n", ((vbe_ib.vbe_version & 0xff00) >> 8), vbe_ib.vbe_version & 0xff);
	clen += sprintf(buf + clen, "Vendor:     %s\n", (char*)vbe_ib.oem_vendor_name_ptr);
	clen += sprintf(buf + clen, "Product:    %s\n", (char*)vbe_ib.oem_product_name_ptr);
	clen += sprintf(buf + clen, "OEM rev:    %s\n", (char*)vbe_ib.oem_product_rev_ptr);
	clen += sprintf(buf + clen, "OEM string: %s\n", (char*)vbe_ib.oem_string_ptr);

	*start = buf + offset;

	if (clen > offset) {
		clen -= offset;
	} else {
		clen = 0;
	}
	return clen;
}

static int __init vesafb_vbe_init(struct fb_info *info)
{
	struct vesafb_task *mytask;
	u16 *mode = 0;
	int i, off = 0;

	vesafb_create_task (mytask);
	mytask->regs.eax = 0x4f00;
	mytask->type = VESAFB_TASK_GETVBE_IB;
	mytask->res = &vbe_ib; 
	vesafb_queue_task (mytask);
	vesafb_wait_for_task(mytask);

	if (vbe_ib.vbe_version < 0x0200) {
		printk(KERN_ERR "vesafb: Sorry, pre-VBE 2.0 cards are not supported.\n");
		kfree(mytask);
		return 1;
	}
	
	if ((mytask->regs.eax & 0xffff) != 0x004f) {
		printk(KERN_ERR "vesafb: Getting mode info block failed (eax=0x%x)\n",
				(u32)mytask->regs.eax);
		kfree(mytask);
		return 1;	
	}
		
	printk(KERN_INFO "vesafb: %s, %s, %s (OEM: %s)\n", (char*)vbe_ib.oem_vendor_name_ptr,
		(char*)vbe_ib.oem_product_name_ptr, (char*)vbe_ib.oem_product_rev_ptr,
		(char*)vbe_ib.oem_string_ptr);

	printk(KERN_INFO "vesafb: VBE version: %d.%d\n",((vbe_ib.vbe_version & 0xff00) >> 8), 
			 vbe_ib.vbe_version & 0xff);

	/* Count available modes. */
	mode = (u16*)vbe_ib.mode_list_ptr;
	while (*mode != 0xffff) {
		vbe_modes_cnt++;
		mode++;
	}

	vbe_modes = kmalloc(sizeof(struct vesafb_mode_info_block)*vbe_modes_cnt,GFP_KERNEL);

	/* Get mode info for all available modes. */
	mode = (u16*)vbe_ib.mode_list_ptr;
	
	while (*mode != 0xffff) {
		mytask->regs.eax = 0x4f01;
		mytask->regs.ecx = (u32) *mode;
		mytask->type = VESAFB_TASK_GETVBE_MODEINFO;
		mytask->res = vbe_modes+off; 
		mytask->done = 0;
		
		vesafb_queue_task (mytask);
		vesafb_wait_for_task(mytask);
	
		vesafb_pmib(mytask->res)->mode_id = *mode;
		
		/* Forget text modes. */
		if ((vesafb_pmib(mytask->res)->mode_attr & 0x10) != 0 && 
		    vesafb_pmib(mytask->res)->bits_per_pixel >= 8) {
			off++;
		} else {
			vbe_modes_cnt--;
		}
		mode++;
	}

	mytask->regs.eax = 0x4f0a;
	mytask->regs.ebx = 0x0;
	mytask->type = VESAFB_TASK_DOVM86;
	mytask->res = NULL;
	mytask->done = 0;

	vesafb_queue_task(mytask);
	vesafb_wait_for_task(mytask);

	if ((mytask->regs.eax & 0xffff) != 0x004f || mytask->regs.es < 0xc000) {
		pmi_setpal = ypan = 0;
	} else {
		pmi_base  = (unsigned short*)phys_to_virt(((unsigned long)mytask->regs.es << 4) + mytask->regs.edi);
		pmi_start = (void*)((char*)pmi_base + pmi_base[1]);
		pmi_pal   = (void*)((char*)pmi_base + pmi_base[2]);
		printk(KERN_INFO "vesafb: protected mode interface info at %04x:%04x\n", 
				 (u16)mytask->regs.es, (u16)mytask->regs.edi);
		printk(KERN_INFO "vesafb: pmi: set display start = %p, set palette = %p\n",
				 pmi_start, pmi_pal);

		if (pmi_base[3]) {
			printk(KERN_INFO "vesafb: pmi: ports = ");
			for (i = pmi_base[3]/2; pmi_base[i] != 0xffff; i++)
				printk("%x ",pmi_base[i]);
			printk("\n");

			if (pmi_base[i] != 0xffff) {
				/*
				 * memory areas not supported (yet?)
				 *
				 * Rules are: we have to set up a descriptor for the requested
				 * memory area and pass it in the ES register to the BIOS function.
				 */				 
				printk(KERN_INFO "vesafb: can't handle memory requests, pmi disabled\n");
				ypan = pmi_setpal = 0;
			}
		}
	}

	if (noedid || vbe_ib.vbe_version < 0x0300)
		goto set_monspecs;
	
	mytask->regs.eax = 0x4f15;
	mytask->regs.ebx = 0;
	mytask->regs.ecx = 0;
	mytask->done = 0; 

	vesafb_queue_task(mytask);
	vesafb_wait_for_task(mytask);

	if ((mytask->regs.eax & 0xffff) != 0x004f)
		goto set_monspecs;

	if ((mytask->regs.ebx & 0x3) == 3) {
		printk(KERN_INFO "vesafb: hardware supports both DDC1 and DDC2 transfers\n");
	} else if ((mytask->regs.ebx & 0x3) == 2) {
		printk(KERN_INFO "vesafb: hardware supports DDC2 transfers\n");
	} else if ((mytask->regs.ebx & 0x3) == 1) {
		printk(KERN_INFO "vesafb: hardware supports DDC1 transfers\n");
	} else {
		printk(KERN_INFO "vesafb: hardware doesn't support DDC transfers\n");
		goto set_monspecs;
	}
	
	mytask->regs.eax = 0x4f15;
	mytask->regs.ebx = 1;
	mytask->regs.ecx = mytask->regs.edx = mytask->done = 0;
	mytask->type = VESAFB_TASK_GETEDID;
	mytask->res = kmalloc(EDID_LENGTH, GFP_KERNEL);

	vesafb_queue_task(mytask);
	vesafb_wait_for_task(mytask);

	if ((mytask->regs.eax & 0xffff) == 0x004f) {
		mon_limits = !fb_get_monitor_limits(mytask->res, &info->monspecs);

		/* If no maximum clock is specified, set to 300 MHz. */
		if (mon_limits && info->monspecs.dclkmax == 0) {
			info->monspecs.dclkmax = 300 * 1000000;
		}

		edid_modes = fb_create_modedb(mytask->res, &edid_modes_cnt);
	}
	kfree(mytask->res);
	
set_monspecs:
	if (maxclk)
		info->monspecs.dclkmax = maxclk * 1000000;

	if (maxvf)
		info->monspecs.vfmax = maxvf;

	if (maxhf)
		info->monspecs.hfmax = maxhf * 1000;

	/* In case DDC transfers are not supported the user can provide monitor limits 
	   manually, lower limits are set to "safe" values. */
	if (!mon_limits && maxclk && maxvf && maxhf) {
		info->monspecs.dclkmin = 0;	
		info->monspecs.vfmin = 60;
		info->monspecs.hfmin = 29000;
	}

	printk(KERN_INFO "vesafb: monitor limits: vf = %d Hz, hf = %d kHz, clk = %d MHz\n", 
		info->monspecs.vfmax, (int)(info->monspecs.hfmax / 1000), 
		(int)(info->monspecs.dclkmax / 1000000));

	kfree(mytask);
	return 0;
}

static int __init decode_mode(u32 *xres, u32 *yres, u32 *bpp, u32 *refresh)
{
	int len = strlen(mode_option), i, err = 0;
	u8 res_specified = 0, bpp_specified = 0, refresh_specified = 0, 
	   yres_specified = 0;

	for (i = len-1; i >= 0; i--) {
 		switch (mode_option[i]) {
		case '@':
    			len = i;
    			if (!refresh_specified && !bpp_specified &&
			    !yres_specified) {
				*refresh = simple_strtoul(&mode_option[i+1], NULL, 0);
				refresh_specified = 1;
			} else
				goto out;
		    	break;
		case '-':
			len = i;
		    	if (!bpp_specified && !yres_specified) {
			    	*bpp = simple_strtoul(&mode_option[i+1], NULL, 0);
				bpp_specified = 1;
		    	} else
				goto out;
		    	break;
		case 'x':
			if (!yres_specified) {
				*yres = simple_strtoul(&mode_option[i+1], NULL, 0);
				yres_specified = 1;
		    	} else
				goto out;
		    	break;
		case '0'...'9':
			break;
		default:
			goto out;
	    	}
	}

	if (i < 0 && yres_specified) {
		*xres = simple_strtoul(mode_option, NULL, 0);
	   	res_specified = 1;
	}

out:	if (!res_specified || !yres_specified) {
		printk(KERN_ERR "vesafb: invalid resolution, %s not specified\n",
				(!res_specified) ? "width" : "height");
		err = -EINVAL;
	}

	return err;
}

static int __init vesafb_init_set_mode(struct fb_info *info)
{
	char buf[32];
	int i, modeid, refresh = 0;
	u8 refresh_specified = 0;

	if (!mode_option)
		mode_option = CONFIG_FB_VESA_DEFAULT_MODE;

	if (vbemode > 0) {
		for (i = 0; i < vbe_modes_cnt; i++) {
			if (vbe_modes[i].mode_id == vbemode) {
				info->var.vmode = FB_VMODE_NONINTERLACED;
				info->var.sync = FB_SYNC_VERT_HIGH_ACT;
				vesafb_setup_var(&info->var, info, &vbe_modes[i]);
				fb_get_mode(FB_VSYNCTIMINGS | FB_IGNOREMON, 60, &info->var, info);
				return i;
			}
		}

		printk(KERN_INFO "specified VBE mode %d not found\n",vbemode);
	}

	if (decode_mode(&info->var.xres, &info->var.yres, 
			&info->var.bits_per_pixel, &refresh)) {
		return -EINVAL;
	}
	if (refresh) {
		refresh_specified = 1;
	} else {
		refresh = 60;
	}
	modeid = vesafb_find_vbe_mode(info->var.xres, info->var.yres, info->var.bits_per_pixel, 0);

	if (modeid == -1) {
		return -EINVAL;	
	} else {
		info->var.vmode = FB_VMODE_NONINTERLACED;
		info->var.sync = FB_SYNC_VERT_HIGH_ACT;
		vesafb_setup_var(&info->var, info, &vbe_modes[modeid]);
	}

	if (gtf)
		goto timings_gtf;

	sprintf(buf, "%dx%d-%d@%d", info->var.xres, info->var.yres, info->var.bits_per_pixel, refresh);
	i = fb_find_mode(&info->var, info, buf, vesa_modes, VESA_MODEDB_SIZE, NULL, 0);
	DPRINTK("fb_find_mode returned %d\n", i);

	if (i == 1 || (i == 2 && vbe_ib.vbe_version < 0x0300))
		return modeid;

	/* The call to fb_find_mode could have changed our var struct, so we need
	 * to restore it. */
	info->var.vmode = FB_VMODE_NONINTERLACED;
	info->var.sync = FB_SYNC_VERT_HIGH_ACT;
	vesafb_setup_var(&info->var, info, &vbe_modes[modeid]);

	if (edid_modes != NULL && !gtf) {
		DPRINTK("looking for EDID modes\n");
		
		for (i = 0; i < edid_modes_cnt; i++) {

			if (edid_modes[i].xres == info->var.xres && 
			    edid_modes[i].yres == info->var.yres && 
			    abs(edid_modes[i].refresh - refresh) < 5) {

				info->var.pixclock = edid_modes[i].pixclock;
			    	info->var.left_margin = edid_modes[i].left_margin;
				info->var.right_margin = edid_modes[i].right_margin;
				info->var.upper_margin = edid_modes[i].upper_margin;
				info->var.lower_margin = edid_modes[i].lower_margin;
				info->var.hsync_len = edid_modes[i].hsync_len;
				info->var.vsync_len = edid_modes[i].vsync_len;
				info->var.sync = edid_modes[i].sync;
				info->var.vmode = edid_modes[i].vmode;
				DPRINTK("using EDID-provided mode\n");
				return modeid;
			}
		}
	}

timings_gtf:
	if (refresh_specified)
		i = FB_VSYNCTIMINGS;
	else
		i = FB_MAXTIMINGS;
		
	if (vbe_ib.vbe_version < 0x0300) { 
		i = FB_VSYNCTIMINGS | FB_IGNOREMON;
		refresh = 60;
	} 
	
	if (!mon_limits)
		i |= FB_IGNOREMON;
	
	if (!fb_get_mode(i, refresh, &info->var, info))
		return modeid;

	printk(KERN_WARNING "vesafb: trying maximum allowed refresh rate\n");

	if (i & FB_VSYNCTIMINGS && refresh > 60 && mon_limits) {
		if (!fb_get_mode(FB_MAXTIMINGS, refresh, &info->var, info))
			return modeid;
	}

	printk(KERN_WARNING "vesafb: using default BIOS refresh rate\n");
	vbemode = vbe_modes[modeid].mode_id;

	return modeid;
}

static int __init vesafb_probe(struct device *device)
{
	char entry[16];
	struct platform_device *dev = to_platform_device(device);
	struct fb_info *info;
	struct vesafb_mode_info_block *mode = NULL;
	int err = 0, i, h;
	unsigned int size_vmode;
	unsigned int size_remap;
	unsigned int size_total;

	vesafb_info = info = framebuffer_alloc(sizeof(struct vesafb_par) + sizeof(u32) * 256, &dev->dev);
	
	if (!info)
	 	return -ENOMEM;

	if (vesafb_pid) {
		vesafb_serv_thread = find_task_by_pid(vesafb_pid);
	} else {
		printk(KERN_ERR "vesafb: vesafb thread not running\n");
		framebuffer_release(info);
		return -EINVAL;
	}

	if (vesafb_vbe_init(info)) {
		printk(KERN_ERR "vesafb: vbe_init failed\n");
		err = -EINVAL;
		goto out;
	}

	vesafb_fix.ypanstep  = ypan     ? 1 : 0;
	vesafb_fix.ywrapstep = (ypan>1) ? 1 : 0;

	info->pseudo_palette = ((u8*)info->par + sizeof(struct vesafb_par));
	info->fbops = &vesafb_ops;
	info->var = vesafb_defined;
	info->fix = vesafb_fix;

	if (fb_alloc_cmap(&info->cmap, 256, 0) < 0) {
		err = -ENXIO;
		goto out;
	}	

	i = vesafb_init_set_mode(info);
	if (i < 0)
		goto out_cmap;

	/* Has the VBE mode number been specified? */
	if (vbemode > 0) {
		for (i = 0; i < vbe_modes_cnt; i++) {
			if (vbe_modes[i].mode_id == vbemode) {
				mode = &vbe_modes[i];
				break;
			}
		}
	} else {
		i = vesafb_find_vbe_mode(info->var.xres, info->var.yres, 
					 info->var.bits_per_pixel, VESAFB_NEED_EXACT_RES | VESAFB_NEED_EXACT_DEPTH);
		if (i > 0)
			mode = &vbe_modes[i];
	}

	/* Find out how much IO memory is required for the mode with
	 * the highest resolution. */
	size_remap = 0;	
	for (i = 0; i < vbe_modes_cnt; i++) {
		h = vbe_modes[i].bytes_per_scan_line * vbe_modes[i].y_res;
		if (h > size_remap)
			size_remap = h;
	}
	size_remap *= 2;
	
	/*   size_vmode -- that is the amount of memory needed for the
	 *                 used video mode, i.e. the minimum amount of
	 *                 memory we need. */
	if (mode != NULL) { 
		size_vmode = info->var.yres * mode->bytes_per_scan_line;
	} else {
		size_vmode = info->var.yres * info->var.xres * 
			     ((info->var.bits_per_pixel + 7) >> 3);
	}

	/*   size_total -- all video memory we have. Used for mtrr
	 *                 entries, ressource allocation and bounds
	 *                 checking. */
	size_total = vbe_ib.total_memory * 65536;
	if (vram_total)
		size_total = vram_total * 1024 * 1024;
	if (size_total < size_vmode)
		size_total = size_vmode;

	/*   size_remap -- the amount of video memory we are going to
	 *                 use for vesafb.  With modern cards it is no
	 *                 option to simply use size_total as that
	 *                 wastes plenty of kernel address space. */
	if (vram_remap)
		size_remap = vram_remap * 1024 * 1024;
	if (size_remap < size_vmode)
		size_remap = size_vmode;
	if (size_remap > size_total)
		size_remap = size_total;

	info->fix.smem_len = size_remap;
	info->fix.smem_start = mode->phys_base_ptr;

	/* We have to set it here, because when setup_var() was called, smem_len
	   wasn't defined yet. */
	info->var.yres_virtual = info->fix.smem_len / mode->bytes_per_scan_line;

	if (ypan && info->var.yres_virtual > info->var.yres) {
		printk(KERN_INFO "vesafb: scrolling: %s using protected mode interface, yres_virtual=%d\n",
		       (ypan > 1) ? "ywrap" : "ypan",info->var.yres_virtual);
	} else {
		printk(KERN_INFO "vesafb: scrolling: redraw\n");
		info->var.yres_virtual = info->var.yres;
		ypan = 0;
	}

	info->flags = FBINFO_FLAG_DEFAULT | 
		(ypan) ? FBINFO_HWACCEL_YPAN : 0;
	
	if (!request_mem_region(info->fix.smem_start, size_total, "vesafb")) {
		printk(KERN_WARNING "vesafb: cannot reserve video memory at 0x%lx\n", info->fix.smem_start);
		/* We cannot make this fatal. Sometimes this comes from magic
		   spaces our resource handlers simply don't know about */
	}
		
	info->screen_base = ioremap(info->fix.smem_start, info->fix.smem_len); 
	
	if (!info->screen_base) {
		printk(KERN_ERR
		       "vesafb: abort, cannot ioremap video memory 0x%x @ 0x%lx\n",
			info->fix.smem_len, info->fix.smem_start);
		err = -EIO;
		goto out_mem;
 	}

	/* Request failure does not faze us, as vgacon probably has this
	   region already (FIXME) */
	request_region(0x3c0, 32, "vesafb");

	if (mtrr && !(info->fix.smem_start & (PAGE_SIZE - 1))) {
		int temp_size = size_total;

		/* Find the largest power-of-two */
		while (temp_size & (temp_size - 1))
                	temp_size &= (temp_size - 1);
                        
                /* Try and find a power of two to add */
		while (temp_size && mtrr_add(info->fix.smem_start, temp_size, MTRR_TYPE_WRCOMB, 1) == -EINVAL) {
			temp_size >>= 1;
		}
	}

	if (register_framebuffer(info) < 0) {
		printk(KERN_ERR "vesafb: failed to register framebuffer device\n");
		err = -EINVAL;
		goto out_mem;
	}

  	printk(KERN_INFO "vesafb: framebuffer at 0x%lx, mapped to 0x%p, using %dk, total %dk\n",
	       info->fix.smem_start, info->screen_base, size_remap/1024, size_total/1024);
	printk(KERN_INFO "fb%d: %s frame buffer device\n", info->node, info->fix.id);

	sprintf(entry, "fb%d", info->node);
	proc_mkdir(entry, 0);

	sprintf(entry, "fb%d/modes", info->node);
	create_proc_read_entry(entry, 0, 0, vesafb_read_proc_modes, NULL);
	
	sprintf(entry, "fb%d/vbe_info", info->node);
	create_proc_read_entry(entry, 0, 0, vesafb_read_proc_vbe_info, NULL);

	fb_destroy_modedb(edid_modes);
	return 0;

out_mem:
	release_mem_region(info->fix.smem_start, size_total);
out_cmap:
	fb_dealloc_cmap(&info->cmap);
out:	
	framebuffer_release(info);
	vesafb_info = NULL;

	fb_destroy_modedb(edid_modes);
	kfree(vbe_modes);
	vbe_modes = NULL;
	return err;
}

int __init vesafb_init(void)
{
	int ret;
#ifndef MODULE
	char *option = NULL;

	if (fb_get_options("vesafb", &option))
		return -ENODEV;
	vesafb_setup(option);
#endif
	ret = driver_register(&vesafb_driver);

	if (!ret) {
		ret = platform_device_register(&vesafb_device);
		if (ret)
			driver_unregister(&vesafb_driver);
	}

	return ret;
}

module_init(vesafb_init);

#ifdef MODULE
void vesafb_exit(void)
{
	char entry[16];

	if (vesafb_info)
		unregister_framebuffer(vesafb_info);

	platform_device_unregister(&vesafb_device);
	driver_unregister(&vesafb_driver);

	if (vesafb_info) {
		sprintf(entry, "fb%d/modes", vesafb_info->node);
		remove_proc_entry(entry, NULL);
	
		sprintf(entry, "fb%d/vbe_info", vesafb_info->node);
		remove_proc_entry(entry, NULL);

		sprintf(entry, "fb%d", vesafb_info->node);
		remove_proc_entry(entry, NULL);

		iounmap(info->screen_base); 
		release_mem_region(vesafb_info->fix.smem_start, vesafb_info->fix.smem_len);
		fb_dealloc_cmap(&vesafb_info->cmap);
		framebuffer_release(vesafb_info);
	}
	
	if (vbe_modes != NULL)
		kfree(vbe_modes);	
}

module_exit(vesafb_exit);

static inline int param_get_scroll(char *buffer, struct kernel_param *kp) { return 0; }
static inline int param_set_scroll(const char *val, struct kernel_param *kp) 
{
	ypan = 0;
	
	if (! strcmp(val, "redraw"))
		ypan=0;
	else if (! strcmp(val, "ypan"))
		ypan=1;
	else if (! strcmp(val, "ywrap"))
		ypan=2;

	return 0;
}

#define param_check_scroll(name, p) __param_check(name, p, void);

module_param_named(scroll, ypan, scroll, 0);
MODULE_PARM_DESC(scroll,"Scrolling mode, set to 'redraw', 'ypan' or 'ywrap'");
module_param_named(vgapal, pmi_setpal, invbool, 0);
MODULE_PARM_DESC(vgapal,"bool: set palette using VGA registers");
module_param_named(pmipal, pmi_setpal, bool, 0);
MODULE_PARM_DESC(pmipal,"bool: set palette using PMI calls");
module_param_named(nomtrr, mtrr, invbool, 0);
MODULE_PARM_DESC(nomtrr,"bool: disable use of MTRR registers");
module_param(nocrtc, bool, 0);
MODULE_PARM_DESC(nocrtc,"bool: ignore CRTC timings when setting modes");
module_param(noedid, bool, 0);
MODULE_PARM_DESC(noedid,"bool: ignore EDID-provided monitor limits when setting modes");
module_param(gtf, bool, 0);
MODULE_PARM_DESC(gtf,"bool: force use of VESA GTF to calculate mode timings");
module_param(vram_remap, uint, 0);
MODULE_PARM_DESC(vram_remap,"Set amount of video memory to be used [MiB]");
module_param(vram_total, uint, 0);
MODULE_PARM_DESC(vram_total,"Set total amount of video memoery [MiB]"); 
module_param(maxclk, ushort, 0);
MODULE_PARM_DESC(maxclk,"Maximum pixelclock [MHz], overrides EDID data");
module_param(maxhf, ushort, 0);
MODULE_PARM_DESC(maxhf,"Maximum horizontal frequency [kHz], overrides EDID data");
module_param(maxvf, ushort, 0);
MODULE_PARM_DESC(maxvf,"Maximum vertical frequency [Hz], overrides EDID data");
module_param_named(mode, mode_option, charp, 0);
MODULE_PARM_DESC(mode,"Specify resolution as \"<xres>x<yres>[-<bpp>][@<refresh>]\"");
module_param(vbemode, ushort, 0);
MODULE_PARM_DESC(vbemode,"VBE mode number to set, overrides 'mode' setting");

#endif /* MODULE */

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Michal Januszewski");
MODULE_DESCRIPTION("Framebuffer driver for VBE2.0+ compliant graphic boards");

